(* mbc.ml :: mbc toplevel :: jec2160 *)
open Ast

let debuglevel = 0									(* 1=debug, 0=no debug *)

(* add variabels to a list, no concern over duplication *)
let add_text variablelist text =
  variablelist := text::!variablelist

(* add de-duplicated variables to a list - use 'add_text_nodup' *)
let rec add_to_list variablelist text = 
  match variablelist with
  | h::t -> if h = text then variablelist else h::add_to_list t text
  | [] -> [text]

let add_text_nodup variablelist text = 
  variablelist := add_to_list !variablelist text

(* top of program *)
let programheader1 =   ref ["#include <stdlib.h>"; "#include <stdio.h>\n"]

(* middle of program, printed after variable list *)
let programheader2 =   ref ["\nvoid main() {\n"]

(* end of program *)
let programfooter1 =   ref ["}\n"]

(* output program body *)
let (programcontents : (string list) ref) = ref []

(* original program body *)
let (originalprogramcontents : (string list) ref) = ref []

(* list of variables that were defined, all are global *)
let (varlist : (string list) ref) = ref []

(* print all variable definitions *)
let printallvariables varlist = List.iter           
  (fun n -> print_string "int ";
            print_string n;
	        print_string ";\n";) varlist
			
(* print list forward *)
let printlistfwd varlist = List.iter (fun n -> print_string n; print_string "\n") !varlist

(* print list backward *)
let printlistback varlist = List.iter (fun n -> print_string n; print_string "\n") (List.rev(!varlist))


	
let rec eval = function 
  | Lit(x) -> string_of_int x              (* number *)
	
  | Id(x) ->                               (* variable *)
      if (debuglevel == 1) then 
	    print_string ("variable name " ^ x ^ "\n") else print_string "";
	  add_text_nodup varlist x;
	  x

  | Endop ->                               (* end of program *)
      if (debuglevel == 1) then
        print_string "endop\n" else print_string "";
	  ""
  
  | Binop(e1, op, e2, e3) ->               (* binary operators *)
      let v1 = eval e1 and v2 = eval e2 and v3 = eval e3 in
      match op with
	  
	  | Add ->                             (* addition *)
		  if (debuglevel == 1) then 
		    print_string "addition\n" else print_string "";      
		  "(" ^ v1 ^ " + " ^ v2 ^ ")"
		  
      | Sub ->                             (* subtraction *)       
		  if (debuglevel == 1) then 
		    print_string "subtraction\n" else print_string "";
		  "(" ^ v1 ^ " - " ^ v2 ^ ")"
		  
      | Mul ->                             (* multiplication *)       
		  if (debuglevel == 1) then 
		    print_string "multiplication\n" else print_string "";
		  "(" ^ v1 ^ " * " ^ v2 ^ ")"
		  
      | Div ->                             (* division *)       
		  if (debuglevel == 1) then 
		    print_string "division\n" else print_string "";
		  "(" ^ v1 ^ " / " ^ v2 ^ ")"
		  
	  | Equal ->                           (* equal *)     
		  if (debuglevel == 1) then 
		    print_string "equal\n" else print_string "";
		  "(" ^ v1 ^ " = " ^ v2 ^ ")"
		  
	  | Notequal ->                        (* not equal *)  
		  if (debuglevel == 1) then 
		    print_string "notequal\n" else print_string "";
		  "(" ^ v1 ^ " != " ^ v2 ^ ")"
		  
	  | Less ->                            (* less than *)      
		  if (debuglevel == 1) then 
		    print_string "less\n" else print_string "";
		  "(" ^ v1 ^ " < " ^ v2 ^ ")"
		  
	  | Greater ->                         (* greater than *)   
		  if (debuglevel == 1) then 
		    print_string "greater\n" else print_string "";
		  "(" ^ v1 ^ " > " ^ v2 ^ ")"
		  
	  | Leq ->                             (* less than or equal *)       
		  if (debuglevel == 1) then 
		    print_string "lessequal\n" else print_string "";
		  "(" ^ v1 ^ " <= " ^ v2 ^ ")"
		  
	  | Geq ->                             (* greater than or equal *)       
		  if (debuglevel == 1) then 
		    print_string "greaterequal\n" else print_string "";
		  "(" ^ v1 ^ " >= " ^ v2 ^ ")"
		  
	  | Ifthenelse ->                      (* if then else *)
		  if (debuglevel == 1) then 
		    print_string "ifthenelse\n" else print_string "";
		  "if " ^ v1 ^ " { " ^ v2 ^ " } else { " ^ v3 ^ " }"
		  
	  | Ifthen ->                          (* if then no else *)
		  if (debuglevel == 1) then 
		    print_string "ifthen\n" else print_string "";
		  "if " ^ v1 ^ " { " ^ v2 ^ " }"
		  
	  | Whileop ->                         (* while do loop *)
		  if (debuglevel == 1) then 
		    print_string "whileop\n" else print_string "";
		  "while " ^ v1 ^ " { " ^ v2 ^ "}"
		  
	  | Forop ->                           (* for loop *)
		  if (debuglevel == 1) then 
		    print_string "forop\n" else print_string "";
		  "for (" ^ v1 ^ ", i < " ^ v2 ^ ", i++) { " ^ v3 ^ " }"

      | Printop ->                         (* printop *)
		  if (debuglevel == 1) then 
		    print_string "printop\n" else print_string "";
		  "printf(" ^ v1 ^ ");"


let _ = 
  (* welcome message *)
  print_endline "\n\nminimalistic basic compiler :: version 1.0 :: joel christner";
  print_endline "project for coms-w4115 :: columbia university :: prof edwards\n";
  print_endline "original program contents";
  print_endline "-------------------------------------------------------------------------------";

  (* gather original program contents *)
  let rec gatherprogram () = 
    try
      let line = input_line stdin in
	    (* print_endline line;      <-- remove for debugging *)
		add_text originalprogramcontents line;
	    gatherprogram ()
	with End_of_file -> 
	  printlistback originalprogramcontents;       (* display originalprogramcontents *)
	  ()
  in
    gatherprogram ();
	
  (* end of gathering program contents *)
  print_endline "\n\n-------------------------------------------------------------------------------";
  print_endline "end of input file, processing...";
  print_endline "-------------------------------------------------------------------------------";
  (* now we have the entire program stored in originalprogramcontents *)

  (* evaluate expressions from each line in originalprogramcontents, store result in programcontents *)
  let programparser = 
    List.iter (fun programline ->   
	            (* print_string ": ";          *)
	            (* print_string programline;   *)             
				(* print_string "\n";          *)
			    let lexbuf = Lexing.from_string programline in                    
				  let expr = Parser.expr Scanner.token lexbuf in
				    let result = eval expr in
				      print_endline result;
					  (* now we have each line evaluated *)
					  (* need to add it to programcontents *)
					  add_text programcontents result
			  ) (List.rev(!originalprogramcontents))
  in
    programparser;
		
  (* final program stored in programcontents - need to display *)
  print_endline "\n\n-------------------------------------------------------------------------------";
  print_endline "output file...";
  print_endline "-------------------------------------------------------------------------------";
  
  (* print programheader1 *)
  printlistfwd programheader1;
  
  (* print list of variables *)
  printallvariables !varlist;
  
  (* print programheader2 *)
  printlistfwd programheader2;
  
  (* print programcontents *)
  printlistback programcontents;
  
  (* print programfooter1 *)
  printlistfwd programfooter1;
  print_endline "\n\n-------------------------------------------------------------------------------";
  print_endline "finished...";
  print_endline "-------------------------------------------------------------------------------";

		
(*
    Open issues in the code...
	- Fix variable declaration issues w/ for, while, if, etc
	  - assign vs compare
	- printop is not evaluating expression before printing
	  - so check associativity and precedence...
	  - check assignment associativity (should be right associative) i.e. a = b = 3
    - comments not being handled properly
	- parsing error when malformed lines are processed - see forloops-bad
	
    Open issues in the project...
	- Test cases and test scripts
	  - integrate tests into makefile
	- Documentation
*)